
/*****************************************************************************
 *
 *	 z80.c
 *	 Portable Z80 emulator
 *
 *	 Copyright (c) 1998 Juergen Buchmueller, all rights reserved.
 *
 *	 - This source code is released as freeware for non-commercial purposes.
 *	 - You are free to use and redistribute this code in modified or
 *	   unmodified form, provided you list me in the credits.
 *	 - If you modify this source code, you must add a notice to each modified
 *	   source file that it has been changed.  If you're a nice person, you
 *	   will clearly mark each change too.  :)
 *	 - If you wish to use this for commercial purposes, please contact me at
 *	   pullmoll@t-online.de
 *	 - The author of this copywritten work reserves the right to change the
 *     terms of its usage and license at any time, including retroactively
 *   - This entire notice must remain in the source code.
 *
 *****************************************************************************/

#include "driver.h"
#include "cpuintrf.h"
#include "z80.h"

#ifndef INLINE
#define INLINE	static __inline
#endif

extern int previouspc;

#define INC_R_ONCE			1	/* compatibility with old Z80 core */
#define BUSY_LOOP_HACKS 	1
#ifndef NO_Z80_BIG_FLAGS_ARRAY
#define BIG_FLAGS_ARRAY 	1	/* big flags array for ADD/ADC/SUB/SBC results */
#else
#define BIG_FLAGS_ARRAY 	0
#endif

#ifdef X86_ASM
#undef  BIG_FLAGS_ARRAY
#define BIG_FLAGS_ARRAY     0
#endif

#include "z80ops.h"

void fake_z80_freemem(void)
{
#if BIG_FLAGS_ARRAY
	if (SZHVC_add) {
		gm_free(SZHVC_add);
		SZHVC_add=NULL;
	}
	if (SZHVC_sub) {
		gm_free(SZHVC_sub);
		SZHVC_sub=NULL;
	}
#endif
}

/****************************************************************************
 * Reset registers to their initial values
 ****************************************************************************/
void Z80_Reset(Z80_DaisyChain *daisy_chain)
{
	int i;
    for (i = 0; i < 256; i++) {
		SZ[i] = 0;
		if (i==0) SZ[i] |= ZF;
		if (i&128) SZ[i] |= SF;
		SZP[i] = SZ[i];
		if ((((i>>7)^(i>>6)^(i>>5)^(i>>4)^(i>>3)^(i>>2)^(i>>1)^i)&1) == 0)
			SZP[i] |= PF;
		SZHV_inc[i] = SZ[i];
		if (i == 0x80) SZHV_inc[i] |= VF;
		if ((i & 0x0f) == 0x00) SZHV_inc[i] |= HF;
        SZHV_dec[i] = SZ[i] | NF;
		if (i == 0x7f) SZHV_dec[i] |= VF;
		if ((i & 0x0f) == 0x0f) SZHV_dec[i] |= HF;
    }
#if BIG_FLAGS_ARRAY
	if (!SZHVC_add || !SZHVC_sub) {
		int oldval, newval;
		UINT8 *padd, *padc, *psub, *psbc;
        /* allocate big flag arrays once */
		SZHVC_add = (UINT8 *)gm_zi_malloc(2*256*256);
		SZHVC_sub = (UINT8 *)gm_zi_malloc(2*256*256);
		if (!SZHVC_add || !SZHVC_sub) {
			return;
		}
		padd = &SZHVC_add[0];
		padc = &SZHVC_add[256*256];
		psub = &SZHVC_sub[0];
		psbc = &SZHVC_sub[256*256];
		for (oldval = 0; oldval < 256; oldval++) {
			for (newval = 0; newval < 256; newval++) {
				/* add or adc w/o carry set */
				*padd = (newval) ? ((newval & 0x80) ? SF : 0) : ZF;
				if ((newval & 0x0f) < (oldval & 0x0f)) *padd |= HF;
				if (newval < oldval) *padd |= CF;
				else if ((newval & 0x80) != (oldval & 0x80)) *padd |= VF;
				padd++;

				/* adc with carry set */
				*padc = (newval) ? ((newval & 0x80) ? SF : 0) : ZF;
                if (((newval-1) & 0x0f) < (oldval & 0x0f)) *padc |= HF;
				if (((newval-1) & 0xff) < oldval) *padc |= CF;
				else if (((newval-1) & 0x80) != (oldval & 0x80)) *padc |= VF;
				padc++;

				/* cp, sub or sbc w/o carry set */
				*psub = NF | ((newval) ? ((newval & 0x80) ? SF : 0) : ZF);
                if ((newval & 0x0f) > (oldval & 0x0f)) *psub |= HF;
				if (newval > oldval) *psub |= CF;
				else if ((newval & 0x80) != (oldval & 0x80)) *psub |= VF;
				psub++;

				/* sbc with carry set */
				*psbc = NF | ((newval) ? ((newval & 0x80) ? SF : 0) : ZF);
                if (((newval+1) & 0x0f) > (oldval & 0x0f)) *psbc |= HF;
				if (((newval+1) & 0xff) > oldval) *psbc |= CF;
				else if (((newval+1) & 0x80) != (oldval & 0x80)) *psbc |= VF;
				psbc++;
			}
		}
	}
#endif
    asm_memset (&Z80, 0, sizeof(Z80_Regs));
	Z80.request_irq = Z80.service_irq = -1;
    change_pc(_PCD);
    _SPD = 0xf000;
#if NEW_INTERRUPT_SYSTEM
	Z80.nmi_state = CLEAR_LINE;
	Z80.irq_state = CLEAR_LINE;
	Z80.pending_irq = 0;
#else
	Z80_Clear_Pending_Interrupts();
#endif
    if (daisy_chain) {
        while( daisy_chain->irq_param != -1 && Z80.irq_max < Z80_MAXDAISY )
        {
            /* set callbackhandler after reti */
			Z80.irq[Z80.irq_max] = *daisy_chain;
            /* device reset */
			if( Z80.irq[Z80.irq_max].reset )
				Z80.irq[Z80.irq_max].reset(Z80.irq[Z80.irq_max].irq_param);
			Z80.irq_max++;
            daisy_chain++;
        }
    }
}

/* search highest interrupt request device (next interrupt device) */
/*    and highest interrupt service device (next reti      device) */
INLINE void check_daisy_chain(void)
{
	int device;
	Z80.request_irq = Z80.service_irq = -1;

	/* search higher IRQ or IEO */
	for( device = 0 ; device < Z80.irq_max ; device ++ ) {
        /* IEO = disable ? */
		if( Z80.int_state[device] & Z80_INT_IEO ) {
			/* if IEO is disable , masking lower IRQ */
			Z80.request_irq = -1;
			/* set highest interrupt service device */
			Z80.service_irq = device;
		}
		/* IRQ = ON ? */
		if( Z80.int_state[device] & Z80_INT_REQ )
			Z80.request_irq = device;
    }
	/* set interrupt pending flag */
	if( Z80.request_irq >= 0 )
		Z80.pending_irq |=	INT_IRQ;
	else
		Z80.pending_irq &= ~INT_IRQ;
}

INLINE void Interrupt(void)
{
	if ( (Z80.pending_irq & NMI_IRQ) || _IFF1 ) {

        int irq_vector = Z80_IGNORE_INT;
	/* there isn't a valid previouspc */
	previouspc = -1;

        /* Check if processor was halted */
	if (_HALT) {
		++_PC;
		_HALT = 0;
	}

        if (Z80.pending_irq & NMI_IRQ) {
		/* Save interrupt flip-flop 1 to 2 */
		_IFF2 = _IFF1;
		/* Clear interrupt flip-flop 1 */
		_IFF1 = 0;
		PUSH( PC );
		_PCD = 0x0066;
		/* reset NMI interrupt request */
		Z80.pending_irq &= ~NMI_IRQ;
	} else {
		/* Clear interrupt flip-flop 1 */
		_IFF1 = 0;
#if NEW_INTERRUT_SYSTEM
		if( !Z80.irq_max ) {  /* not daisy chain mode */
			/* call back the cpu interface to retrieve the vector */
			Z80.vector = (*Z80.irq_callback)(0);
		}
#else
			/* reset INT interrupt request */
		Z80.pending_irq &= ~INT_IRQ;
#endif
		if ( Z80.irq_max ) {
                	if( Z80.request_irq >= 0 ) {
                    		irq_vector = Z80.irq[Z80.request_irq].interrupt_entry(Z80.irq[Z80.request_irq].irq_param);
                    		Z80.request_irq = -1;
               		}
            	} else {
                	irq_vector = Z80.vector;
           	}
            	/* Interrupt mode 2. Call [Z80.I:databyte] */
		if( _IM == 2 ) {
			irq_vector = (irq_vector & 0xff) | (_I << 8);
			PUSH( PC );
			_PCD = RM16( irq_vector );
		} else
			/* Interrupt mode 1. RST 38h */
			if( _IM == 1 ) {
				/* R_INC; wasn't there... */
				EXEC(op,0xff);
			} else {
				/* Interrupt mode 0. We check for CALL and JP instructions, */
				/* if neither of these were found we assume a 1 byte opcode */
				/* was placed on the databus								*/
				switch (irq_vector & 0xff0000) {
					case 0xcd0000:	/* call */
						PUSH( PC );
					case 0xc30000:	/* jump */
						_PCD = irq_vector & 0xffff;
						break;
					default:
						irq_vector &= 0xff;
						/* R_INC; wasn't there... */
						EXEC(op,irq_vector);
						break;
				}
			}
		}
    }
    change_pc(_PCD);
}

#if NEW_INTERRUPT_SYSTEM

void Z80_set_nmi_line(int state)
{
    if (Z80.nmi_state == state) return;
    Z80.nmi_state = state;
	if (state != CLEAR_LINE) {
        Z80.pending_irq |= NMI_IRQ;
	}
}

void Z80_set_irq_line(int irqline, int state)
{
    Z80.irq_state = state;
	if (state == CLEAR_LINE) {
		if( !Z80.irq_max ) {
			if(!_IFF1 ) 						/* if interrupts are disabled */
				Z80.pending_irq &= ~INT_IRQ;	/* clear the pending IRQ */
		}
	}  else {
		if( Z80.irq_max ) {
			int daisychain, dc_device, dc_state;
			daisychain = (*Z80.irq_callback)(irqline);
			dc_device = daisychain >> 8;
			dc_state = daisychain & 0xff;
			if( Z80.int_state[dc_device] != dc_state ) {
				/* set new interrupt status */
				Z80.int_state[dc_device] = dc_state;
				/* check interrupt status */
				check_daisy_chain();
			}
		} else {
			if( _IFF1 ) 					/* if interrupts are enabled */
				Z80.pending_irq |= INT_IRQ; /* set the pending IRQ */
		}
	}
}

void Z80_set_irq_callback(int (*callback)(int))
{
    Z80.irq_callback = callback;
}

#else

void Z80_Cause_Interrupt(int type)
{
/* type value :                                                          */
/*  Z80_NMI_INT                      -> NMI request                      */
/*  Z80_IGNORE_INT                   -> no request                       */
/*  vector(0x00-0xff)                -> SINGLE interrupt request         */
/*  Z80_VECTOR(device,status)        -> DaisyChain change interrupt status */
/*      device : device number of daisy-chain link                       */
/*      status : Z80_INT_REQ  -> interrupt request                       */
/*               Z80_INT_IEO  -> interrupt disable output                */

    if (type == Z80_NMI_INT) {
        Z80.pending_irq |= NMI_IRQ;
	} else if (type != Z80_IGNORE_INT) {
        if( Z80.irq_max ) {   /* daisy chain mode */
            int device = type >> 8;
            int state  = type & 0xff;

            if( Z80.int_state[device] != state ) {
                /* set new interrupt status */
                Z80.int_state[device] = state;
                /* check interrupt status */
                check_daisy_chain();
            }
        } else {
            /* single int mode */
			Z80.vector = type & 0xff;
            Z80.pending_irq |= INT_IRQ;
        }
    }
}

void Z80_Clear_Pending_Interrupts(void)
{
    int i;

    /* clear irq for all device */
    for( i = 0 ; i < Z80_MAXDAISY ; i++ )
    {
        /* !!!!! shoud be clear interrupt status for daisy-chain devices !!!!! */
        Z80.int_state[i]  = 0;
    }
    Z80.pending_irq = 0;
	Z80.service_irq = -1;
}

#endif

/****************************************************************************
 * Set all registers to given values
 ****************************************************************************/
void Z80_SetRegs (Z80_Regs *Regs)
{
	Z80 = *Regs;
	change_pc(_PCD);
}

/****************************************************************************
 * Get all registers in given buffer
 ****************************************************************************/
void Z80_GetRegs (Z80_Regs *Regs)
{
	*Regs = Z80;
}

/****************************************************************************
 * Return program counter
 ****************************************************************************/
unsigned Z80_GetPC (void)
{
	return _PCD;
}

/****************************************************************************
 * Execute IPeriod T-states. Return number of T-states really executed
 ****************************************************************************/
int Z80_Execute(int cycles)
{
	Z80_ICount = cycles;
	do {
		if (Z80.pending_irq)
			Interrupt();
		R_INC;
        	previouspc = _PCD;
		EXEC(op,ROP());
	} while (Z80_ICount > 0);

	return cycles - Z80_ICount;
}
